/**
 * Flext external code for Max/MSP and PureData
 * Leslie cabinet model presented in the DAFx-11 paper
 *
 * version 0.1
 *
 * Created by Jussi Pekonen
 **/

/* Required libraries */
#include <flext.h> // For building compatible externals for both Max/MSP and Pd
#include <cstdlib> // For the memory allocation functions

/** The SDF class **/
class SDF: public flext_dsp {
	
	/* The Flext header */
    FLEXT_HEADER(SDF, flext_dsp)
    
	/* Public methods and variables of the class */
	public:
		// Constructor
		SDF(int argc, const t_atom *argv);
		// Destructor
		~SDF() {
			free(lesliedelayline);
		}
	
	/* Protected methods and variables of the class */
	protected:
		// DSP calculation
		virtual void m_signal(int n, float *const *in, float *const *out);
	
	/* Private methods and variables of the class */
	private:
		// The delay line
		float* lesliedelayline;
		// Delay line length
		int lesliedllength;
		// The number of allpass filters
		int leslieallpasscount;
		// The function to change the allpass chain output position
		FLEXT_CALLBACK_1(setAllpassChainLength, float);
		void setAllpassChainLength(float length);
};

/** Associate the external to an object name **/
FLEXT_NEW_DSP_V("SDF~ leslie~", SDF)

/** Implementation of the class constructor **/
SDF::SDF(int argc, const t_atom *argv) {
	
	/* Check the creation parameters */
	lesliedllength = 16;
	leslieallpasscount = 5;
	switch (argc) {
		case 2:
			lesliedllength = (IsFloat(argv[1])) ? ((int) GetFloat(argv[1])) : (IsInt(argv[1])) ? (GetInt(argv[1])) : 16;
		case 1:
			leslieallpasscount = (IsFloat(argv[0])) ? ((int) GetFloat(argv[0])) : (IsInt(argv[0])) ? (GetInt(argv[0])) : 5;
			break;
		default:
			lesliedllength = 16;
			leslieallpasscount = 5;
			break;
	}
	
	/* Initialize the Leslie parameters */
	// Check that the number of allpass filters does not exceed the number of delays
	lesliedllength = (leslieallpasscount > lesliedllength) ? (leslieallpasscount + 5) : lesliedllength;
	// Initialize the delay line with zeros
	lesliedelayline = (float*) malloc(lesliedllength*sizeof(float));
	int k;
	for (k = 0; k < lesliedllength; k++)
		lesliedelayline[k] = 0.0;

	/* Inlets and outlets */
	// Inlets
	AddInSignal("Input signal");
	AddInSignal("Modulation signal");
	AddInFloat("The number of allpass filters in the chain (positive integer)");
	// Outlets
	AddOutSignal("The Leslie tone");

	/* Assign functions for inlet 3 */
	FLEXT_ADDMETHOD(2, setAllpassChainLength);

}

/** Implementation of the setAllpassChainLength function **/
void SDF::setAllpassChainLength(float length) {
	// Set the allpass chain length
	leslieallpasscount = (length > lesliedllength) ? lesliedllength : length;
}

/** Implementation of the virtual DSP function **/
void SDF::m_signal(int n, float *const *in, float *const *out) {
	/* Variables */
	// The input signal inlet pointer
	float* input = in[0];
	// The modulator signal inlet pointer
	float* modulator = in[1];
	// The output signal outlet pointer
	float* output = out[0];
	
	// Signal and coefficient variables
	float filtin, filtout, outvalue, coeff;
	// Looping variable
	int k;
	
	/* Compute the output */
	// Generate the required number of samples
	while (n--) {
		// Take the input sample
		filtin = *input++; outvalue = filtin;
		// Take the coefficient value
		coeff = *modulator++;
		// Loop through the allpass chain
		for (k = 0; k < lesliedllength; k++) {
			// Compute the input signal for the next allpass filter
			filtout = -coeff*filtin + (1.0 + coeff)*lesliedelayline[k];
			// Compute the delay line value and update it
			lesliedelayline[k] = (1.0 - coeff)*filtin + coeff*lesliedelayline[k];
			// Set the input sample for the next allpass filter
			filtin = filtout;
			if (k == leslieallpasscount - 1)
				outvalue = filtout;
		}
		// Output the correct sample
		*output++ = outvalue;
	}
}
